@@ -20,8 +20,9 @@ def masked_mean(t, mask, dim = 1):
 
 # sampling helpers
 
-def top_k(logits, thres = 0.9):
-    k = int((1 - thres) * logits.shape[-1])
+def top_k(logits, thres = 0.5):
+    num_logits = logits.shape[-1]
+    k = max(int((1 - thres) * num_logits), 1)
     val, ind = torch.topk(logits, k)
     probs = torch.full_like(logits, float('-inf'))
     probs.scatter_(1, ind, val)
@@ -34,16 +35,16 @@ def generate_images(
     text,
     clipper = None,
     mask = None,
-    filter_thres = 0.9,
+    filter_thres = 0.5,
     temperature = 1.
 ):
     x = text
-    out = x
 
     text_seq_len = model.text_seq_len
     image_seq_len = model.image_seq_len
     total_len = text_seq_len + model.image_seq_len - text.shape[1]
 
+    out = x
     for _ in range(total_len):
         text, image = x[:, :text_seq_len], x[:, text_seq_len:]
         logits = model(text, image, mask = mask)[:, -1, :]
@@ -57,14 +58,15 @@ def generate_images(
             mask = F.pad(mask, (0, 1), value=True)
 
     text_seq = torch.cat((x[:, :1], out[:, :(text_seq_len - 1)]), dim = 1)
-    img_seq = out[:, -(image_seq_len + 1):-1]
+
+    img_seq = out[:, -image_seq_len:]
     img_seq -= model.num_text_tokens
-    img_seq.clamp_(min = 0, max = (model.num_image_tokens - 1)) # extra insurance - todo: get rid of this at a future date and rely only on masking of logits
 
     images = vae.decode(img_seq)
 
     if exists(clipper):
-        scores = clipper(text_seq, img_seq, return_loss = False)
+        print(images.shape)
+        scores = clipper(text_seq, images, return_loss = False)
         return images, scores
 
     return images
@@ -74,10 +76,11 @@ def generate_images(
 class DiscreteVAE(nn.Module):
     def __init__(
         self,
-        num_tokens,
+        num_tokens = 512,
         dim = 512,
         hidden_dim = 64,
-        num_layers = 3
+        num_layers = 3,
+        channels = 3
     ):
         super().__init__()
         hdim = hidden_dim
@@ -87,7 +90,7 @@ class DiscreteVAE(nn.Module):
         encoder_layers = []
         decoder_layers = []
         for i in range(num_layers):
-            enc_in = 3 if i == 0 else hdim
+            enc_in = channels if i == 0 else hdim
             encoder_layers += [
                 nn.Conv2d(enc_in, hdim, 4, stride = 2, padding = 1),
                 nn.ReLU(),
@@ -100,7 +103,7 @@ class DiscreteVAE(nn.Module):
             ]
             
         encoder_layers.append(nn.Conv2d(hdim, num_tokens, 1))
-        decoder_layers.append(nn.Conv2d(hdim, 3, 1))
+        decoder_layers.append(nn.Conv2d(hdim, channels, 1))
 
         self.encoder = nn.Sequential(*encoder_layers)
         self.decoder = nn.Sequential(*decoder_layers)
@@ -156,13 +159,15 @@ class CLIP(nn.Module):
         dim_image = 512,
         dim_latent = 512,
         num_text_tokens = 10000,
-        num_visual_tokens = 512,
         text_enc_depth = 6,
-        visual_enc_depth = 6,
         text_seq_len = 256,
-        visual_seq_len = 1024,
         text_heads = 8,
+        num_visual_tokens = 512,
+        visual_enc_depth = 6,
         visual_heads = 8,
+        visual_image_size = 256,
+        visual_patch_size = 32,
+        channels = 3,
         vae = None
     ):
         super().__init__()
@@ -171,8 +176,13 @@ class CLIP(nn.Module):
         self.text_transformer = Encoder(dim = dim_text, depth = text_enc_depth, heads = text_heads)
         self.to_text_latent = nn.Linear(dim_text, dim_latent, bias = False)
 
-        self.visual_emb = nn.Embedding(num_visual_tokens, dim_image)
-        self.visual_pos_emb = nn.Embedding(visual_seq_len, dim_image)
+        assert visual_image_size % visual_patch_size == 0, 'Image dimensions must be divisible by the patch size.'
+        num_patches = (visual_image_size // visual_patch_size) ** 2
+        patch_dim = channels * visual_patch_size ** 2
+
+        self.visual_patch_size = visual_patch_size
+        self.to_visual_embedding = nn.Linear(patch_dim, dim_image)
+        self.visual_pos_emb = nn.Embedding(num_patches, dim_image)
         self.visual_transformer = Encoder(dim = dim_image, depth = visual_enc_depth, heads = visual_heads)
         self.to_visual_latent = nn.Linear(dim_image, dim_latent, bias = False)
 
@@ -190,7 +200,7 @@ class CLIP(nn.Module):
         text_mask = None,
         return_loss = False
     ):
-        b, device = text.shape[0], text.device
+        b, device, p = text.shape[0], text.device, self.visual_patch_size
 
         if exists(self.vae):
             image = self.vae.get_codebook_indices(image)
@@ -198,9 +208,9 @@ class CLIP(nn.Module):
         text_emb = self.text_emb(text)
         text_emb += self.text_pos_emb(torch.arange(text.shape[1], device = device))
 
-
-        image_emb = self.visual_emb(image)
-        image_emb += self.visual_pos_emb(torch.arange(image.shape[1], device = device))
+        image_patches = rearrange(image, 'b c (h p1) (w p2) -> b (h w) (p1 p2 c)', p1 = p, p2 = p)
+        image_emb = self.to_visual_embedding(image_patches)
+        image_emb += self.visual_pos_emb(torch.arange(image_emb.shape[1], device = device))
 
         enc_text = self.text_transformer(text_emb, mask = text_mask)
         enc_image = self.visual_transformer(image_emb)
@@ -253,7 +263,10 @@ class DALLE(nn.Module):
         self.num_image_tokens = num_image_tokens
         self.text_seq_len = text_seq_len
         self.image_seq_len = image_seq_len
-        self.total_tokens = num_text_tokens + num_image_tokens + 1 # extra for EOS
+
+        seq_len = text_seq_len + image_seq_len
+        total_tokens = num_text_tokens + num_image_tokens + 1 # extra for EOS
+        self.total_tokens = total_tokens
         
         self.vae = vae
         if exists(self.vae):
@@ -267,6 +280,20 @@ class DALLE(nn.Module):
             nn.Linear(dim, self.total_tokens),
         )
 
+        seq_range = torch.arange(seq_len)
+        logits_range = torch.arange(total_tokens)
+
+        seq_range = rearrange(seq_range, 'n -> () n ()')
+        logits_range = rearrange(logits_range, 'd -> () () d')
+
+        logits_mask = (
+            ((seq_range >= (text_seq_len - 1)) & (logits_range < num_text_tokens)) |
+            ((seq_range < (text_seq_len - 1)) & (logits_range >= num_text_tokens)) |
+            ((seq_range != (seq_len - 1)) & (logits_range >= (total_tokens - 1)))
+        )
+
+        self.register_buffer('logits_mask', logits_mask)
+
     def forward(
         self,
         text,
@@ -280,15 +307,19 @@ class DALLE(nn.Module):
         tokens = self.text_emb(text)
         tokens += self.text_pos_emb(torch.arange(text.shape[1], device = device))
 
+        seq_len = tokens.shape[1]
+
         if exists(image) and not is_empty(image):
             is_raw_image = len(image.shape) == 4
+            image_len = image.shape[1]
+            seq_len += image_len
 
             if is_raw_image:
                 assert exists(self.vae), 'VAE must be passed into constructor if you are to train directly on raw images'
                 image = self.vae.get_codebook_indices(image)
 
             image_emb = self.image_emb(image)
-            image_emb += self.image_pos_emb(torch.arange(image.shape[1], device = device))
+            image_emb += self.image_pos_emb(torch.arange(image_len, device = device))
 
             tokens = torch.cat((tokens, image_emb), dim = 1)
 
@@ -298,6 +329,11 @@ class DALLE(nn.Module):
         out = self.transformer(tokens, mask = mask)
         logits = self.to_logits(out)
 
+        # mask logits to make sure text predicts text (except last token), and image predicts image
+        mask = self.logits_mask[:, :seq_len]
+        max_neg_value = -torch.finfo(logits.dtype).max
+        logits.masked_fill_(mask, max_neg_value)
+
         if not return_loss:
             return logits
 
